# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301  USA

from wcs.qommon.humantime import *

from wcs.workflows import Workflow, WorkflowStatusItem, register_item_class, \
        render_list_of_roles, get_role_translation

from wcs.formdata import Evolution
from qommon.form import *
from qommon.publisher import get_publisher_class
from qommon.cron import CronJob

class TimeoutWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Timeout')
    key = 'timeout'
    endpoint = False

    status = None
    timeout = None

    _granularity = 20 * 60 # default: 20 minutes, see bottom of file

    def render_as_line(self):
        status = [ x for x in self.parent.parent.possible_status \
                   if str(x.id) == str(self.status)]
        if status and self.timeout:
            return _('Timeout to "%(status)s" after %(time)s') % {'status': status[0].name, 'time': seconds2humanduration(self.timeout)}
        else:
            return _('Timeout (not completed)')

    def fill_admin_form(self, form):
        form.add(SingleSelectWidget, 'status', title = _('Status'), value = self.status,
                options = [(None, '---')] + [(x.id, x.name) for x in self.parent.parent.possible_status])
        _hint = _("You can use the following variables: %(variables)s, "\
                  "but minimal duration is %(granulariy)s") % {
                        'variables': ','.join(timewords()),
                        'granularity': seconds2humanduration(self._granularity)}
        form.add(StringWidget, 'timeout', title = _('Timeout'), value = seconds2humanduration(self.timeout), hint = _hint )

    def submit_admin_form(self, form):
        widget = form.get_widget('timeout')
        if widget:
            try:
                self.timeout = int(widget.parse())
            except:
                self.timeout = humanduration2seconds(widget.parse())
        for f in ('status',):
            widget = form.get_widget(f)
            if widget:
                setattr(self, f, widget.parse())


def workflows_with_timeout():
    '''Return a list of workflow objects with at least a status with a timeout object
    '''
    wfs_status = {}

    for id in Workflow.keys():
        workflow = Workflow.get(id, ignore_errors = True)
        if not workflow:
            continue
        for status in workflow.possible_status:
            for item in status.items:
                if isinstance(item, TimeoutWorkflowStatusItem) and item.status and item.timeout:
                    if not wfs_status.get(id):
                        wfs_status[id] = {}
                    status = 'wf-%s' % status.id
                    if not wfs_status[id].get(status):
                        wfs_status[id][status] = []
                    wfs_status[id][status].append(item)

    return wfs_status

def _apply_timeouts(publisher):
    '''Traverse all filled form and apply expired timeout jumps if needed'''
    from formdef import FormDef
    wfs_status = workflows_with_timeout()

    for formdef_id in FormDef.keys():
        formdef = FormDef.get(formdef_id, ignore_errors = True)
        if not formdef:
            continue
        if str(formdef.workflow_id) in wfs_status.keys():
            for formdata_id in formdef.data_class().keys():
                formdata = formdef.data_class().get(formdata_id, ignore_errors = True)
                if not formdata:
                    continue
                if formdata.status in wfs_status[str(formdef.workflow_id)]:
                    if formdata.evolution:
                        last = formdata.evolution[-1].time
                    else:
                        last = formdata.receipt_time
                    diff = time.time() - time.mktime(last)
                    for x in wfs_status[str(formdef.workflow_id)][formdata.status]:
                        if diff > x.timeout:
                            evo = Evolution()
                            evo.time = time.localtime()
                            evo.status = 'wf-%s' % x.status
                            if not formdata.evolution:
                                formdata.evolution = []
                            formdata.evolution.append(evo)
                            formdata.status = evo.status
                            formdata.store()
                            formdata.perform_workflow()
                            break

register_item_class(TimeoutWorkflowStatusItem)

if get_publisher_class():
    # every 20 minutes check for expired status jump timeouts. If you change
    # cronjob schedule, change also the granularity variable
    TimeoutWorkflowStatusItem._granularity = 20*60 # 20 minutes
    get_publisher_class().register_cronjob(
            CronJob(_apply_timeouts, hours=range(24), minutes=range(0,60,20)))
